const DB = (() => {

  /** 
   * @private
   * The specific title of DB class errors
   * @constant string 
   */
  const errorTitle = "DB ";

  /** 
   * @private
   * ID of db conncetion session
   * @var String
   */  
  var session_id;

  /** 
   * @private
   * Default URL for db connection
   * @var String 
   */  
  var url;

  /** 
   * @private
   * New instance of @class Log
   * @constant Object
   */ 
  const log = new Log();

  /** 
   * @private
   * New instance of @class Spinal
   * @constant Object
   */ 
  const spnl = new Spinal();

  /** 
   * @private
   * Instance of the TableDOM class
   * @var Object 
   */
  var that;

  /** 
   * @private
   * @enum Storage types codes 
   */  
  const strgCodes = {
    s: window.sessionStorage,
    l: window.localStorage
  }

  /** 
   * @private
   * @enum Codes of commands for Storage types
   */  
  const commands = ['s', 'g', 'r', 'c', 'l'];

  /** 
   * @public
   * Status of the current db session
   * @var boolean 
   */  
  DB.prototype.sessionIsset = false;

  /** 
   * @public
   * Status of login part
   * @var boolean 
   */  
  DB.prototype.loggedin = false;

  /** 
   * @public
   * @constructor
   * @param customUrl   Custom connection URL (default = null)
   */
  function DB (customUrl = null)
  {
    that = this;
    this.data = {};
    const sessionInfo = this.checkSession();

    // Create the DB request url root if the session information are set
    if (sessionInfo !== false)
    {
      url = "http://" + sessionInfo.ip + ":" + sessionInfo.port + "/json/" + sessionInfo.db + "/";
    }
    else
    {
      url = false;
      if (customUrl !== null)
      {
        if (log.checkType({"s:customUrl":customUrl}) === false) return;
        this.customUrl = customUrl;

        spnl.execute();
      }
    }
  }

  /** 
   * @public
   * Storage controller for both Local and Session storages
   * @param method    Storage and command codes
   * @param key       The key name of the storage item (default = null)
   * @param val       The value of the storage item (default = null)
   * @returns
   */  
  DB.prototype.store = (method, key = null, val = null) => {
    if (log.checkType({"s:method":method}) === false) return;
    try
    {
      if (!method.includes(':'))
        throw new SyntaxError("The method must contain ':' between storage code and command code!");
      
      if (method.length > 3)
        throw new SyntaxError("Unexpected characters!");

      const mtdCodes = method.split(':');
      const strg = strgCodes[mtdCodes[0]];
      if (typeof strg === "undefined")
        throw new SyntaxError("Undefined storage code!");
      
      const cmd = mtdCodes[1];
      if (commands.indexOf(cmd) === -1)
        throw new SyntaxError("Undefined command code!");
      
      if(typeof key !== "string") key = JSON.stringify(key);
      if(typeof val !== "string") val = JSON.stringify(val);

           if (cmd === 's') return strg.setItem(key, val);
      else if (cmd === 'g') return strg.getItem(key);
      else if (cmd === 'r') return strg.removeItem(key);
      else if (cmd === 'c') return strg.clear();
      else if (cmd === 'l') return strg.length;
    }
    catch (error)
    {
      console.error(errorTitle + error);
    }
  }

  /** 
   * @public
   * Check Session information in storages
   * @returns User information
   */  
  DB.prototype.checkSession = () => {
    // 'l' => localStorage, 's' => sessionStorage
    const lIp   = that.store("l:g", "ip");
    const lPort = that.store("l:g", "port");
    const lDB   = that.store("l:g", "db");
    const sIp   = that.store("s:g", "ip");
    const sPort = that.store("s:g", "port");
    const sDB   = that.store("s:g", "db");

    if (!(!lIp || lIp == '') && !(!lPort || lPort == '') && !(!lDB || lDB == ''))
      return { ip: lIp, port: lPort, db: lDB };
    
    else if (!(!sIp || sIp == '') && !(!sPort || sPort == '') && !(!sDB || sDB == ''))
      return { ip: sIp, port: sPort, db: sDB };
    
    return false;
  }

  /** 
   * @private
   * Check Session information in storages
   * @returns User information
   */ 
  function checkUser ()
  {
    const lUser = that.store("l:g", "user");
    const lPass = that.store("l:g", "password");
    const sUser = that.store("s:g", "user");
    const sPass = that.store("s:g", "password");

    if (!(!lUser || lUser == '') && !(!lPass || lPass == ''))
      return { user: lUser, pass: lPass };
    
    else if (!(!sUser || sUser == '') && !(!sPass || sPass == ''))
      return { user: sUser, pass: sPass };
    
    return false;
  }

  /** 
   * @public
   * Open new db session
   * @param check   If false return when the url = false (default = false)
   */  
  DB.prototype.open = (check = false) => {
    if (log.checkType({"b:check":check}) === false) return;
    try
    {
      if (check === false)
        if (url === false)
        {
          alert("Incorrect connection settings!");
          settingsPage(true);
          return;
        }
      
      if (that.store("l:g", "lastSession") === null)
        $.ajax({
          url: url + "open",
          dataType: "JSON",
          timeout: 5000,
          error: (jqXHR, exception) => {
            ajaxError(jqXHR, exception, msg => {
              alert(msg + " Your connection settings may be incorrect.");
              settingsPage(true);
            });
            return;
          },
          success: (data) => {
            if(typeof data.session === "undefined")
            {
              DB.prototype.sessionIsset = false;
              alert("The app failed to start! You may have no license.");
              window.location.reload();
            }
            session_id = data.session;

            that.store("l:s", "lastSession", data.session);

            that.company = data.CompanyName;
            that.dataName = data.dataName;

            DB.prototype.sessionIsset = true;

            spnl.execute();
          }
        });
      else
      {
        DB.prototype.sessionIsset = true;
        spnl.execute();
      }
    }
    catch (error)
    {
      console.error(errorTitle + error);
      return false;
    }
  }

  /** 
   * @private
   * Get response status and display errors
   * @param jqXHR       jQuery XMLHttpRequest object
   * @param exception   The request exception
   * @param callback    Callback a function after checking status (optional)
   * @returns Error message
   */
  function ajaxError (jqXHR, exception, callback = null)
  {
    let msg = '';
    if (jqXHR.status === 0) msg = 'Not connect Verify Network.';
    else if (jqXHR.status == 404) msg = 'Requested page not found. [404]';
    else if (jqXHR.status == 500) msg = 'Internal Server Error [500].';
    else if (exception === 'parsererror') msg = 'Requested JSON parse failed.';
    else if (exception === 'timeout') msg = 'Time out error.';
    else if (exception === 'abort') msg = 'Ajax request aborted.';
    else msg = 'Uncaught Error.\n' + jqXHR.responseText;
    if (callback !== null)
    {
      if (log.checkType({"f:callback":callback}) === false) return msg;
      callback(msg);
    }
    return msg;
  }

  /** 
   * @public
   * Log in with user information
   */  
  DB.prototype.login = (username = null, password = null, errorCallback = null) => {
    try
    {
      if (DB.prototype.sessionIsset === false)
      {
        DB.prototype.store("l:s", "loggedin", false);
        throw new Error("Session ID not found");
      }

      if (DB.prototype.store("l:g", "loggedin") == "true") {
        spnl.execute();
        return;
      }

      let command;
      const userInfo = checkUser();
      const lastSession = that.store("l:g", "lastSession");
      const validSession = lastSession === null ? session_id : lastSession;
      if (userInfo === false)
      {
        if (username === null || password === null)
        {
          loginPage(that.login);
          
          return;
        }
        else if (log.checkType({"s:username":username, "s:password":password}) === false) return;

        command = url + "login?session=" + validSession + "&user=" + username + "&password=" + password;
      }
      else command = url + "login?session=" + validSession + "&user=" + userInfo.user + "&password=" + userInfo.pass;

      $.ajax({
        url: command,
        dataType: "JSON",
        timeout: 5000,
        error: (jqXHR, exception) => {
          ajaxError(jqXHR, exception, msg => {
            DB.prototype.store("l:s", "loggedin", false);
            alert(msg + " Your connection settings may be incorrect.");
            settingsPage(true);
          });
          return;
        },
        success: data => {
          if (errorCallback !== null)
            errorCallback(data);
          if (typeof data.Session !== "undefined") {
            DB.prototype.store("l:s", "loggedin", true);
            DB.prototype.store("l:r", "password");
            DB.prototype.store("s:r", "password");
          } else {
            DB.prototype.store("l:s", "loggedin", false);
            if (confirm("Something went wrong. Would you like to log in again?"))
            {
              that.close();
              that.store("l:r", "lastSession");
              window.location.reload();
            }
            throw new Error("Failed to login!");
          }
          spnl.execute();
        }
      });
    }
    catch (error)
    {
      console.error(errorTitle + error);
      return false;
    }
  }

  /** 
   * @public
   * Close the current session
   * @param msg   Message to alert after closing session (Optional)
   */  
  DB.prototype.close = (msg = null) => {
    if (msg !== null)
      if (log.checkType({"s:msg":msg}) === false) return;
    try
    {
      $.ajax({
        url: url + "close?session=" + that.store("l:g", "lastSession"),
        dataType: "JSON",
        timeout: 5000,
        error: (jqXHR, exception) => {
          alert(ajaxError(jqXHR, exception));
        },
        success: () => {
          DB.prototype.loggedin = false;
          if (msg !== null) alert(msg);
          spnl.execute();
        }
      });
    }
    catch (error)
    {
      console.error(errorTitle + error);
      return false;
    }
  }

  /** 
   * @public
   * Special DB query method
   * @param table    Table name
   * @param params   query parameters
   * @retruns loadJSON void
   */  
  DB.prototype.query = (table, params = null, callback = null) => {
    if (log.checkType({"s:table":table}) === false) return;
    if (params !== null)
      if (log.checkType({"a:params":params}) === false) return;
    
    try
    {
      let parameters = "";
      if (params !== null) parameters = "&" + [...params].toString().replace(/,/g, "&");
      const commandString = url + table + "?session=" + that.store("l:g", "lastSession") + parameters;
      $.ajax({
        url: commandString,
        dataType: "JSON",
        timeout: 5000,
        error: (jqXHR, exception) => {
          alert(jqXHR);
          alert(exception);
          ajaxError(jqXHR, exception, msg => {
            alert(msg + " Your connection settings may be incorrect.");
            settingsPage(true);
          });
          return;
        },
        success: data => {
          if (data.error_message === "Session not found")
          {
            that.store("l:r", "lastSession");
            that.store("l:r", "loggedin");
            window.location.reload();
          }

          if (callback === null) return data;

          else if (typeof callback === "function") {
            callback(data);
            spnl.execute();
          }
        }
      });
    }
    catch (error)
    {
      console.error(errorTitle + error);
      return false;
    }
  }

  DB.prototype.cacheAccounts = () => {
    that.query("accounts", null, data => {
      that.data["accounts"] = [];
      let count = 0;
      let length = 0;
      for (let i = 0; i < data[data.length - 1].ElmID + 1; i++)
      {
        if (typeof data[i - count] !== "undefined" && data[i - count].ElmID == i) {
          if (data[i - count].ElmIsBook === false) {
            that.data["accounts"].push(data[i - count]);
            length++;
          }
          else that.data["accounts"].push(-1);
        } else {
          that.data["accounts"].push(-1);
          count++;
        }
      }
      that.store("s:s", "accountsLength", length);
    });
  }

  DB.prototype.cacheMaterials = () => {
    that.query("materials", null, data => {
      that.data["materials"] = [];
      let count = 0;
      let length = 0;
      for (let i = 0; i < data[data.length - 1].ElmID + 1; i++)
      {
        if (typeof data[i - count] !== "undefined" && data[i - count].ElmID == i) {
          if (data[i - count].ElmIsBook === false) {
            that.data["materials"].push(data[i - count]);
            length++;
          }
          else that.data["materials"].push(-1);
        } else {
          that.data["materials"].push(-1);
          count++;
        }
      }
      that.store("s:s", "materialsLength", length);
    });
  }

  DB.prototype.warehouses = () => {
    that.query("warehouses", null, data => {
      that.data["warehouses"] = new Map();
      for (let i = 0; i < data.length; i++)
        that.data["warehouses"].set(data[i].ElmID, data[i].ElmName);
    });
  }
  return DB;
})();